--[[
AdiFocusBar - Smart focus bar for hunters.
Copyright 2010-2013 Adirelle (adirelle@gmail.com)
All rights reserved.
--]]

if select(2, UnitClass("player")) ~= "HUNTER" then return end
local addonName, ns = ...
local addon, Debug = ns.addon, ns.Debug

------------------------------------------------------------------------------
-- Constants
------------------------------------------------------------------------------

local BAR_BACKDROP = {
	bgFile = [[Interface\Addons\AdiFocusBar\media\white16x16]],
	tile = true,
	tileSize = 16,
}

local BORDER_SIZE = 2
local BORDER_BACKDROP = {
	edgeFile = [[Interface\Addons\AdiFocusBar\media\white16x16]],
	edgeSize = BORDER_SIZE,
}

local BAR_TEXTURE = [[Interface\TargetingFrame\UI-StatusBar]]

local LSM_STATUSBAR = "BantoBar"

local WIDTH, HEIGHT = 250, 15

------------------------------------------------------------------------------
-- Marks
------------------------------------------------------------------------------

local function Mark_SetPosition(mark, value, text, icon)
	if value then
		local bar = mark:GetParent()
		local min, max = bar:GetMinMaxValues()
		if value > min and value < max then
			local valueToPixel = bar:GetWidth() / (max - min)
			mark:SetPoint("CENTER", bar, "LEFT", valueToPixel * (value - min), 0)
			mark:Show()
			local t = mark.Text
			if t then
				t:SetShown(not not text)
				if text then
					t:SetText(text)
				end
			end
			local i = mark.Icon
			if i then
				i:SetShown(not not icon)
				if icon then
					i:SetTexture(select(3, GetSpellInfo(icon)))
				end
			end
			return
		end
	end
	mark:Hide()
end

local function SpawnMark(bar, r, g, b)
	local mark = CreateFrame("Frame", nil, bar)
	mark:SetSize(2, HEIGHT)
	mark.SetPosition = Mark_SetPosition

	local line = mark:CreateTexture(nil, "BACKGROUND")
	line:SetPoint("CENTER")
	line:SetSize(2, HEIGHT)
	line:SetTexture([[Interface\Addons\AdiFocusBar\media\white16x16]])
	line:SetVertexColor(r, g, b)

	local text = mark:CreateFontString(nil, "OVERLAY")
	text:SetPoint("CENTER")
	local name, height, flags = _G.GameFontWhiteSmall:GetFont()
	text:SetFont(name, height, "OUTLINE")
	text:Hide()
	mark.Text = text

	local icon = mark:CreateTexture(nil, "ARTWORK")
	icon:SetPoint("CENTER")
	icon:SetSize(HEIGHT * 0.66, HEIGHT * 0.66)
	icon:Hide()
	mark.Icon = icon

	return mark
end

------------------------------------------------------------------------------
-- Bar
------------------------------------------------------------------------------

local function Bar_SetValues(bar, newValue, newMax, newMin)
	if newValue and newMax then
		local value, min, max = bar:GetValue(), bar:GetMinMaxValues()
		if min ~= (newMin or 0) or max ~= newMax then
			bar:SetMinMaxValues((newMin or 0), newMax)
		end
		if value ~= newValue then
			bar:SetValue(newValue)
		end
		if not bar:IsShown() then
			bar:Show()
		end
	elseif bar:IsShown() then
		bar:Hide()
	end
end

local function Bar_FixTexture(bar)
	local texture = bar:GetStatusBarTexture()
	local value, min, max = bar:GetValue(), bar:GetMinMaxValues()
	if max ~= min then
		texture:SetTexCoord(0, (value-min)/(max-min), 0, 1)
	end
end

local function Bar_OnValueChanged(bar)
	local value = bar:GetValue()
	addon.Text:SetFormattedText("%d", value)
	bar.Spark:SetPosition(value)
	Bar_FixTexture(bar)
end

local function SpawnSpark(bar)
	local spark = bar:CreateTexture(nil, "OVERLAY")
	spark:SetTexture([[Interface\CastingBar\UI-CastingBar-Spark]])
	spark:SetBlendMode('ADD')
	spark:SetSize(20, HEIGHT*2.5)
	spark.SetPosition = Mark_SetPosition
	return spark
end

------------------------------------------------------------------------------
-- Spell forecast
------------------------------------------------------------------------------

local function ForeCast_UpdateSpell(icon)
	local r, g, b = 1, 1, 1
	if icon.spell then
		if IsSpellInRange(icon.spellName) == 0 then
			r, g, b = 1, 0.3, 0.3, 1
		else
			local usable, nomana = IsUsableSpell(icon.spell)
			if not usable then
				if nomana then
					r, g, b = 0.3, 0.3, 1
				else
					r, g, b = 0.5, 0.5, 0.5
				end
			end
		end
	end
	icon:SetVertexColor(r, g, b, 1)
end

local function Forecast_Update(self, event)
	if self.spell then
		CooldownFrame_SetTimer(self.Cooldown, GetSpellCooldown(self.spell))
	else
		self.Cooldown:Hide()
	end
	self.timer = 0
end

local function Forecast_OnShow(self)
	self:RegisterEvent('SPELL_UPDATE_COOLDOWN')
	self:RegisterEvent('SPELL_UPDATE_USABLE')
	self:Update('OnShow')
end

local function Forecast_OnHide(self)
	self:UnregisterAllEvents()
end

local function Forecast_OnUpdate(self, elapsed)
	if self.fadeInIcon then
		local alpha = self.fadeInIcon:GetAlpha()
		if alpha < 1 then
			alpha = min(alpha + elapsed / 0.4, 1)
			self.fadeInIcon:SetAlpha(alpha)
			self.fadeOutIcon:SetAlpha(1-alpha)
		else
			self.fadeInIcon = nil
		end
	end
	if self.timer > 0 then
		self.timer = self.timer - elapsed
	else
		self.timer = 0.1
		ForeCast_UpdateSpell(self.Icon1)
		ForeCast_UpdateSpell(self.Icon2)
	end
end

local function Forecast_SetSpell(self, spell)
	if spell == self.spell then return end
	self.spell = spell
	local icon1, icon2 = self.Icon1, self.Icon2

	if spell ~= icon1.spell and spell ~= icon2.spell then
		local newIcon = icon1:GetAlpha() < icon2:GetAlpha() and icon1 or icon2
		newIcon:SetTexture(spell and select(3, GetSpellInfo(spell)) or "")
		newIcon.spell = spell
		newIcon.spellName = spell and GetSpellInfo(spell)
	end

	if icon1.spell == spell and icon1:GetAlpha() < 1 then
		self.fadeInIcon, self.fadeOutIcon = icon1, icon2
	elseif icon2.spell == spell and icon2:GetAlpha() < 1 then
		self.fadeInIcon, self.fadeOutIcon = icon2, icon1
	else
		self.fadeInIcon, self.fadeOutIcon = nil, nil
	end

	self:Update("SetSpell", spell)
end

------------------------------------------------------------------------------
-- Setup
------------------------------------------------------------------------------

local function Frame_Enable(self)
	Debug(self, 'Frame_Enable')
	addon.db.profile.enabled[self.name] = true
	self:Show()
end

local function Frame_Disable(self)
	Debug(self, 'Frame_Disable')
	addon.db.profile.enabled[self.name] = false
	self:Hide()
end

local function Frame_IsEnabled(self)
	Debug(self, 'Frame_IsEnabled', addon.db.profile.enabled[self.name])
	return addon.db.profile.enabled[self.name]
end

local function SpawnFrameWithBorder(width, height, name, label, ...)
	local frame = CreateFrame("Frame", addonName..name, addon)
	frame.name = name
	frame:SetSize(width, height)
	frame:SetBackdrop(BAR_BACKDROP)
	frame:SetBackdropColor(0, 0, 0, 0.8)
	frame:SetBackdropBorderColor(0, 0, 0, 0)

	local border = CreateFrame("Frame", nil, frame)
	border:SetSize(width+BORDER_SIZE*2, height+BORDER_SIZE*2)
	border:SetPoint("CENTER")
	border:SetBackdrop(BORDER_BACKDROP)
	border:SetBackdropColor(0, 0, 0, 0)
	border:SetBackdropBorderColor(0, 0, 0, 1)
	frame.Border = border

	frame:SetPoint(...)

	frame.Enable = Frame_Enable
	frame.Disable = Frame_Disable
	frame.IsEnabled = Frame_IsEnabled
	frame.LM10_Enable = Frame_Enable
	frame.LM10_Disable = Frame_Disable
	frame.LM10_IsEnabled = Frame_IsEnabled

	addon:RegisterMovable(frame, function()
		return addon.db.profile.layout[name]
	end, addonName..' '..label, border)

	frame:SetShown(frame:IsEnabled())

	return frame
end

function addon:SetupGUI()
	local bar = SpawnFrameWithBorder(WIDTH, HEIGHT, "Bar", "bar", "BOTTOM", self, 0, 208)
	self.Bar = bar

	local focusBar = CreateFrame("StatusBar", nil, bar)
	focusBar:SetAllPoints()
	focusBar:SetSize(WIDTH - HEIGHT - 2, HEIGHT)
	focusBar:SetStatusBarTexture(BAR_TEXTURE)
	focusBar:SetScript('OnMinMaxChanged', Bar_OnValueChanged)
	focusBar:SetScript('OnValueChanged', Bar_OnValueChanged)
	focusBar:SetScript('OnShow', Bar_OnValueChanged)
	focusBar.SetValues = Bar_SetValues
	focusBar.Spark = SpawnSpark(focusBar)
	self.FocusBar = focusBar

	local fillerBar = CreateFrame("StatusBar", nil, bar)
	fillerBar:SetAllPoints(focusBar)
	fillerBar:SetStatusBarTexture(BAR_TEXTURE)
	fillerBar:SetScript('OnMinMaxChanged', Bar_FixTexture)
	fillerBar:SetScript('OnValueChanged', Bar_FixTexture)
	fillerBar.SetValues = Bar_SetValues
	self.FillerBar = fillerBar

	focusBar:SetFrameLevel(fillerBar:GetFrameLevel()+1)

	local lsm = LibStub('LibSharedMedia-3.0', true)
	if lsm then
		local function Update(value)
			local texture = lsm:Fetch("statusbar", value, true)
			focusBar:SetStatusBarTexture(texture)
			fillerBar:SetStatusBarTexture(texture)
		end
		Update(LSM_STATUSBAR)

		lsm.RegisterCallback(focusBar, 'LibSharedMedia_SetGlobal', function(_, _, media, value)
			if media == "statusbar" then
				Update(value)
			end
		end)
	end

	focusBar:SetStatusBarColor(PowerBarColor.FOCUS.r, PowerBarColor.FOCUS.g, PowerBarColor.FOCUS.b)
	fillerBar:SetStatusBarColor(PowerBarColor.FOCUS.b, PowerBarColor.FOCUS.r, PowerBarColor.FOCUS.g)
	fillerBar:Hide()

	local text = focusBar:CreateFontString(nil, "OVERLAY", "GameFontWhite")
	text:SetAllPoints(focusBar)
	text:SetJustifyH("RIGHT")
	text:SetJustifyV("MIDDLE")
	text:SetShadowColor(0,0,0,1)
	text:SetShadowOffset(1, -1)
	self.Text = text

	self.Marks = {
		ArcaneShot = SpawnMark(focusBar, 1, 0.8, 0.8),
		BlackArrow = SpawnMark(focusBar, 1, 0, 1),
		ExplosiveShot = SpawnMark(focusBar, 1, 0.7, 0),
		SerpentSting = SpawnMark(focusBar, 0, 1, 0),
	}
	self.Marks.KillCommand = self.Marks.ExplosiveShot
	self.Marks.ChimeraShot = self.Marks.ExplosiveShot

	local forecast = SpawnFrameWithBorder(HEIGHT, HEIGHT, "ForeCast", "forecast", "RIGHT", bar, "LEFT", -BORDER_SIZE, 0)
	forecast.timer = 0
	forecast.Update = Forecast_Update
	forecast.SetSpell = Forecast_SetSpell
	forecast:SetScript('OnShow', Forecast_OnShow)
	forecast:SetScript('OnHide', Forecast_OnHide)
	forecast:SetScript('OnEvent', Forecast_Update)
	forecast:SetScript('OnUpdate', Forecast_OnUpdate)
	self.Forecast = forecast

	local cooldown = CreateFrame("Cooldown", nil, forecast)
	cooldown:SetAllPoints()
	forecast.Cooldown = cooldown

	local icon1 = forecast:CreateTexture(nil, "ARTWORK")
	icon1:SetSize(HEIGHT, HEIGHT)
	icon1:SetPoint("CENTER")
	icon1:SetTexCoord(0.05, 0.95, 0.05, 0.95)
	forecast.Icon1 = icon1

	local icon2 = forecast:CreateTexture(nil, "ARTWORK")
	icon2:SetSize(HEIGHT, HEIGHT)
	icon2:SetPoint("CENTER")
	icon2:SetTexCoord(0.05, 0.95, 0.05, 0.95)
	icon2:SetAlpha(0)
	forecast.Icon2 = icon2

	self:SetScript('OnShow', self.OnShow)
	self:SetScript('OnHide', self.OnHide)
	self:SetScript('OnUpdate', self.OnUpdate)

	self:UpdateMovableLayout()
end
