--[[
AdiFocusBar - Smart focus bar for hunters.
Copyright 2010-2013 Adirelle (adirelle@gmail.com)
All rights reserved.
--]]

if select(2, UnitClass("player")) ~= "HUNTER" then return end
local addonName, ns = ...
local addon, Debug = ns.addon, ns.Debug

------------------------------------------------------------------------------
-- Logic
------------------------------------------------------------------------------

local BASE_REGEN = 4.0
local FILLER_GAIN = 14
local GCD_DURATION = 1

-- Upvalues
local fillerShot
local now, hasteFactor = GetTime(), 1
local focus, focusMax = 0, 100
local targetHealthRatio = 1
local steadiesInRow = 0
local fillerCastingTime = 2
local currentCast, currentCastEndTime = nil, 0
local lastCastSpell, lastCastTime = nil, 0

local EstimatePassiveIncome

local zeroMT = { __index = function(t,k) t[k] = 0 return 0 end }
local _cooldowns = setmetatable({}, zeroMT)
local _auras = setmetatable({}, zeroMT)
local costs = setmetatable({}, zeroMT)
local known = {}

local cooldowns = setmetatable({}, { __index = function(t, k) return max(0, _cooldowns[k] - now) end})
local auras = setmetatable({}, { __index = function(t, k) return max(0, _auras[k] - now) end})

local function EstimatePassiveIncome_Basic(delay)
	if not delay or delay < 0 then return 0 end
	return BASE_REGEN * hasteFactor * delay + ceil(min(delay, auras.Fervor) / 2) * 5
end

local function EstimatePassiveIncome_ViperVenom(delay)
	if not delay or delay < 0 then return 0 end
	return EstimatePassiveIncome_Basic(delay) + ceil(min(delay, auras.SerpentSting) / 3) * 3
end

local function EstimatePassiveIncome_RapidRecuperation(delay)
	if not delay or delay < 0 then return 0 end
	return EstimatePassiveIncome_Basic(delay) + ceil(min(delay, auras.RapidFire) / 3) * 12
end

-- S = [S]pell ids
local S = setmetatable({
	AimedShot       =  19434,
	AncientHysteria =  90355,
	ArcaneShot      =   3044,
	BestialWrath    =  19574,
	BlackArrow      =   3674,
	CarefulAim      =  34483,
	ChimeraShot     =  53209,
	CobraShot       =  77767,
	Crows           = 131894,
	DireBeast       = 120679,
	ExplosiveShot   =  53301,
	Fervor          =  82726,
	FocusFire       =  82692,
	Frenzy          =  19623,
	GlaiveToss      = 117050,
	KillCommand     =  34026,
	KillShot        =  53351,
	PowerShot       = 109259,
	Rabid           =  53401,
	RapidFire       =   3045,
	Readiness       =  23989,
	SerpentSting    =   1978,
	Stampede        = 121818,
	SteadyFocus     =  53224,
	SteadyShot      =  56641,
}, { __index = function(t, k)
	if k ~= nil then
		error("Unlisted spell: "..k)
	end
end})

-- N = spell [N]ames
local N = setmetatable({}, {__index = function(t, k)
	local v = k and GetSpellInfo(S[k]) or nil
	t[k] = v
	return v
end})

local outOfGCD = {
	[S.Readiness] = true,
	[S.Fervor] = true,
	[S.RapidFire] = true,
	[S.BestialWrath] = true,
	[S.Rabid] = true,
	[S.AncientHysteria] = true,
}

local function GetCastingTime(id)
	if not id or outOfGCD[id] then
		return 0
	else
		return max((select(7, GetSpellInfo(id)) or 0) / 1000, GCD_DURATION)
	end
end

local function GetCost(spell)
	if spell == S.CobraShot or spell == S.SteadyShot or spell == N.CobraShot or spell == N.SteadyShot then
		return -FILLER_GAIN
	elseif spell == S.Fervor or spell == N.Fervor then
		return -50
	elseif spell then
		return select(4, GetSpellInfo(spell)) or 0
	end
end

local function BuildAuraGetter(unit, name)
	local filter = (unit == "target") and "PLAYER|HARMFUL" or "PLAYER|HELPFUL"
	return function()
		local expires = select(7, UnitAura(unit, N[name], nil, filter))
		_auras[name] = type(expires) == "number" and expires or 0
	end
end

local function BuildCooldownGetter(name)
	local id = S[name]
	return function()
		local start, duration, enable = GetSpellCooldown(id)
		if enable == 1 and start and duration and duration > GCD_DURATION then
			_cooldowns[name] = start + duration
		else
			_cooldowns[name] = 0
		end
	end
end

local GetSpellDefs = function()
	local spells = {}

	local spellMeta = { __index = {
		Aura = function(self, unit, ...)
			local num = select('#', ...)
			if num == 0 then return self:Aura(unit, self.name) end
			if not self.aura then
				self.aura = {}
			end
			for i = 1, num do
				local name = select(i, ...)
				self.aura[name] = BuildAuraGetter(unit, name)
			end
			return self
		end,
		Cost = function(self, spell)
			local name = spell or self.name
			local id = S[name]
			self.cost = function()
				costs[name] = select(4, GetSpellInfo(id)) or 0
			end
			return self
		end,
		Cooldown = function(self, ...)
			local num = select('#', ...)
			if num == 0 then return self:Cooldown(self.name) end
			if not self.cooldown then
				self.cooldown = {}
			end
			for i = 1, num do
				local name = select(i, ...)
				self.cooldown[name] = BuildCooldownGetter(name)
			end
			return self
		end,
		UseASAP = function(self)
			local name, id = self.name, S[self.name]
			if self.cost and self.cooldown and self.cooldown[name] then
				return self:UseWhen(function() return costs[name], cooldowns[name], id end)
			elseif self.cost then
				return self:UseWhen(function() return costs[name], 0, id end)
			elseif self.cooldown and self.cooldown[name] then
				return self:UseWhen(function() return 0, cooldowns[name], id end)
			else
				return self:UseWhen(function() return 0, 0, id end)
			end
			return self
		end,
		UseWhen = function(self, f)
			self.when = f
			return self
		end
	}}

	local function Spell(name, actualSpell)
		local spell = setmetatable({ name = actualSpell or name }, spellMeta)
		if name == '*' then
			spell.isAvailable = function() return true end
		else
			local id = S[spell.name]
			spell.isAvailable = function() return IsPlayerSpell(id) or IsSpellKnown(id) or IsSpellKnown(id, true)  end
		end
		spells[name] = spell
		return spell
	end

	Spell('*')
		:Aura('player', 'Fervor', 'RapidFire')
		:Aura('target', 'SerpentSting')

	Spell('KillShot'):Cooldown():UseWhen(function()
		if targetHealthRatio <= 0.2 then
			return 0, cooldonws.KillShot, S.KillShot
		end
	end)

	Spell('BlackArrow'):Cooldown():Cost():Aura('target'):UseWhen(function()
		local timeLeft = max(0, auras.BlackArrow-1)
		if lastCastSpell == S.BlackArrow then
			timeLeft = lastCastTime + 20
		end
		return costs.BlackArrow, max(timeLeft, cooldowns.BlackArrow), S.BlackArrow
	end)

	Spell('KillCommand'):Cooldown():Cost():UseASAP()

	Spell('ChimeraShot'):Cooldown():Cost():UseASAP()

	Spell('ExplosiveShot'):Cooldown():Cost():UseASAP()

	Spell('SerpentSting'):Aura('target'):Cost():UseWhen(function()
		local timeLeft = max(0, auras.SerpentSting - 1)
		if lastCastSpell == S.SerpentSting or (lastCastSpell == S.ChimeraShot and lastCastTime < timeLeft) then
			timeLeft = lastCastTime + 15 - now
		elseif lastCastSpell == S.CobraShot and lastCastTime + fillerCastingTime < timeLeft then
			timeLeft = timeLeft + 6
		elseif currentCast == S.CobraShot and currentCastEndTime < timeLeft then
			timeLeft = timeLeft + 6
		end
		return costs.SerpentSting, timeLeft, S.SerpentSting
	end)

	Spell('SteadyFocus'):Aura('player'):UseWhen(function()
		local required = 2 - steadiesInRow
		if currentCast == N.SteadyShot then
			required = required + 1
		end
		if required > 0 then
			return -FILLER_GAIN, max(0, auras.SteadyFocus - fillerCastingTime * required - 1), S.SteadyShot
		end
	end)

	Spell('GlaiveToss'):Cooldown():Cost():UseASAP()

	Spell('PowerShot'):Cooldown():Cost():UseASAP()

	Spell('AimedShot'):Cooldown():Cost():UseWhen(function()
		if costs.AimedShot == 0 or targetHealthRatio >= 0.8 then
			return costs.AimedShot, 0, S.AimedShot
		end
	end)

	Spell('CarefulSteadyShot', 'CarefulAim'):UseWhen(function()
		if targetHealthRatio >= 0.8 then
			return -FILLER_GAIN, 0, S.SteadyShot
		end
	end)

	Spell('ArcaneShot'):Cost():UseASAP()

	Spell('SteadyShot'):UseWhen(function() return -FILLER_GAIN, 0, S.SteadyShot end)

	Spell('CobraShot'):UseWhen(function() return -FILLER_GAIN, 0, S.CobraShot end)

	Spell('FocusFire'):Aura('player', 'Frenzy', 'FocusFire'):Aura('pet', 'BestialWrath'):UseWhen(function()
		local stack = select(4, UnitAura("player", N.Frenzy, nil, "HELPFUL|PLAYER")) or 0
		if stack < 5 then return end
		return 0, max(
			min(auras.FocusFire, auras.Frenzy),
			auras.BestialWrath
		), S.FocusFire
	end)

	Spell('Fervor'):Aura('player'):Cooldown():UseWhen(function()
		if focus < 40 and UnitPower("pet", SPELL_POWER_FOCUS) < 50 then
			return -50, max(cooldowns.Fervor, auras.Fervor), S.Fervor
		end
	end)

	Spell('DireBeast'):Cooldown():UseWhen(function()
		return 0, max(cooldowns.DireBeast, auras.DireBeast), S.DireBeast
	end)

	Spell('BestialWrath'):Cooldown('BestialWrath', 'KillCommand'):Aura('pet', 'BestialWrath'):UseWhen(function()
		return 60, max(cooldowns.BestialWrath, cooldowns.KillCommand, auras.BestialWrath), S.BestialWrath
	end)

	Spell('Stampede'):Cooldown('Stampede', 'RapidFire'):UseWhen(function()
		if auras.RapidFire == 0 and cooldowns.RapidFire < 60 then
			return 0, max(cooldowns.Stampede, cooldowns.RapidFire) , S.Stampede
		else
			return 0, cooldowns.Stampede, S.Stampede
		end
	end)

	Spell('RapidFire'):Cooldown():Aura('player'):UseWhen(function()
		return 0, max(cooldowns.RapidFire, auras.RapidFire), S.RapidFire
	end)

	Spell('AncientHysteria'):Cooldown():Aura('pet'):UseWhen(function()
		return 0, max(cooldowns.AncientHysteria, auras.AncientHysteria), S.AncientHysteria
	end)

	Spell('Rabid'):Cooldown():Aura('pet', 'Rabid', 'BestialWrath'):UseWhen(function()
		return 0, max(cooldowns.BestialWrath, cooldowns.Rabid, auras.Rabid), S.Rabid
	end)

	Spell('Readiness'):Cooldown():UseWhen(function()
		if cooldowns.RapidFire > 30
			and (not known.BestialWrath or cooldowns.BestialWrath > 0)
			and (not known.Crows or cooldowns.Crows > 0)
			and (not known.DireBeast or cooldowns.DireBeast > 0)
			and (not known.KillCommand or cooldowns.KillCommand > 0)
			and (not known.ChimeraShot or cooldowns.ChimeraShot > 0)
			and (not known.GlaiveToss or cooldowns.GlaiveToss > 0)
		then
			return 0, cooldowns.Readiness, S.Readiness
		end
	end)

	Spell('Crows'):Cost():Cooldown():Aura('target'):UseWhen(function()
		return costs.Crows, max(cooldowns.Crows, auras.Crows), S.Crows
	end)

	GetSpellDefs = function() return spells end

	return spells
end

local rotation = { rotations = {} }
local rotationsBySpec = {
	-- Beast Mastery
	[253] = {
		selector = function() return auras.BestialWrath > 0 and "BestialWrath" or "default" end,
		rotations = {
			default = {
				"SerpentSting",
				"KillShot",
				"BestialWrath",
				"Rabid",
				"KillCommand",
				"GlaiveToss",
				"PowerShot",
				"Crows",
				"DireBeast",
				"FocusFire",
				"AncientHysteria",
				"RapidFire",
				"Stampede",
				"ArcaneShot",
				"Fervor",
				"Readiness",
				"CobraShot"
			},
			BestialWrath = {
				"KillShot",
				"Rabid",
				"KillCommand",
				"GlaiveToss",
				"Crows",
				"ArcaneShot",
				"Fervor",
				"CobraShot"
			},
		},
	},
	-- Marksmanship
	[254] = {
		selector = function()
			return known.CarefulAim and targetHealthRatio >= 0.8 and "CarefulAim" or "default"
		end,
		rotations = {
			default = {
				"SerpentSting",
				"SteadyFocus",
				"ChimeraShot",
				"RapidFire",
				"AncientHysteria",
				"Rabid",
				"Stampede",
				"KillShot",
				"AimedShot",
				"CarefulSteadyShot",
				"GlaiveToss",
				"PowerShot",
				"Crows",
				"DireBeast",
				"ArcaneShot",
				"Fervor",
				"Readiness",
				"SteadyShot"
			},
			CarefulAim = {
				"SerpentSting",
				"SteadyFocus",
				"ChimeraShot",
				"RapidFire",
				"AimedShot",
				"SteadyShot",
			},
		}
	},
	-- Survival
	[255] = {
		selector = function() return "default" end,
		rotations = {
			default = {
				"SerpentSting",
				"BlackArrow",
				"KillShot",
				"ExplosiveShot",
				"GlaiveToss",
				"PowerShot",
				"Crows",
				"DireBeast",
				"RapidFire",
				"AncientHysteria",
				"Rabid",
				"Stampede",
				"ArcaneShot",
				"Fervor",
				"Readiness",
				"CobraShot"
			}
		},
	}
}

local updaters = {
	costs = {},
	cooldowns = {},
	auras = {}
}
local calculate = {}

local updateTimer = 0
local spec

function IncludeSpell(defs, spell)
	local def = defs[spell]
	if not def then
		error("Unknwon spell "..spell.." in rotation "..rotationName.." for "..spec)
	elseif def.isAvailable() then
		if spell ~= '*' then
			if not known[spell] then
				Debug(spell, 'known')
				known[spell] = S[spell]
				calculate[spell] = def.when
			end
		end
		if def.cost then
			Debug('Watching', spell, 'cost')
			updaters.costs[spell] = def.cost
		end
		if def.cooldown then
			for k, v in pairs(def.cooldown) do
				if not updaters.cooldowns[k] then
					Debug('Watching', k, 'cooldown')
					updaters.cooldowns[k] = v
				end
			end
		end
		if def.aura then
			for k, v in pairs(def.aura) do
				if not updaters.auras[k] then
					Debug('Watching aura', k)
					updaters.auras[k] = v
				end
			end
		end
		return true
	else
		Debug('Unknown spell: ', spell, 'ignored')
	end
end

function addon:SPELLS_CHANGED(event)
	local specGroup = GetSpecialization()
	spec = specGroup and GetSpecializationInfo(specGroup)
	if not spec then return end

	-- Cleanup previous data
	wipe(known)
	wipe(calculate)
	wipe(updaters.costs)
	wipe(updaters.cooldowns)
	wipe(updaters.auras)
	wipe(rotation.rotations)

	local defs = GetSpellDefs()

	-- Some spell should always be watched
	IncludeSpell(defs, '*')

	-- Build rotation order depending on available spells
	local currentRotation = rotationsBySpec[spec]
	rotation.selector = currentRotation.selector
	for rotationName, spells in pairs(currentRotation.rotations) do
		local actualRotation = {}
		rotation.rotations[rotationName] = actualRotation
		for rank, spell in ipairs(spells) do
			if IncludeSpell(defs, spell) then
				tinsert(actualRotation, spell)
			end
		end
		Debug(rotationName, 'rotation:', unpack(actualRotation))
	end

	-- Estimating income
	if known.SerpentSting and IsPlayerSpell(118976) then
		EstimatePassiveIncome = EstimatePassiveIncome_ViperVenom
	elseif known.RapidFire and IsPlayerSpell(53232) then
		EstimatePassiveIncome = EstimatePassiveIncome_RapidRecuperation
	else
		EstimatePassiveIncome = EstimatePassiveIncome_Basic
	end

	fillerShot = known.CobraShot or known.SteadyShot

	self:UNIT_AURA(event, "player")
	self:SPELL_UPDATE_COOLDOWN(event)

	updateTimer = 0
end
addon.PLAYER_TALENT_UPDATE = addon.SPELLS_CHANGED

local function SpellForecast(prediction, beforeNextAction)
	local candidate, candidateCost, candidateTime
	for i, name in ipairs(rotation.rotations[rotation.selector()]) do
		local cost, timeLeft, spell = calculate[name]()
		if cost and timeLeft then
			local startTime, endTime, startFocus, endFocus
			if outOfGCD[spell] and not currentCast then
				startTime, endTime = timeLeft, timeLeft
				startFocus = focus + EstimatePassiveIncome(startTime)
				endFocus = startFocus - cost
			else
				startTime = max(timeLeft, beforeNextAction)
				endTime = startTime + GetCastingTime(spell)
				startFocus = prediction + EstimatePassiveIncome(startTime)
				endFocus = prediction - cost + EstimatePassiveIncome(endTime)
			end
			if startFocus < cost and startFocus + FILLER_GAIN >= cost then
				startTime = max(beforeNextAction, timeLeft - fillerCastingTime)
				endTime = startTime + fillerCastingTime + GetCastingTime(spell)
				spell, cost = fillerShot, cost - FILLER_GAIN
				startFocus = prediction + EstimatePassiveIncome(startTime)
				endFocus = prediction - cost + EstimatePassiveIncome(endTime)
			end
			if startFocus >= cost and endFocus < focusMax then
				if not candidate then
					candidate, candidateCost, candidateTime = spell, cost, startTime
				elseif startTime < candidateTime and endTime <= candidateTime + GCD_DURATION * 0.66 then
					local newCandidateTime = max(endTime, candidateTime)
					local newCandidateFocus = prediction - cost + EstimatePassiveIncome(newCandidateTime)
					if newCandidateFocus >= candidateCost then
						candidate, candidateCost, candidateTime = spell, cost, startTime
					end
				end
			end
		end
	end
	return candidate
end

function addon:Update(event)
	-- Refresh some upvalues
	now = GetTime()
	hasteFactor = 1.0 + GetRangedHaste() / 100
	fillerCastingTime = max(select(7, GetSpellInfo(fillerShot)) / 1000,  GCD_DURATION)

	-- Update the focus upvalues
	focus, focusMax = UnitPower("player", SPELL_POWER_FOCUS), UnitPowerMax("player", SPELL_POWER_FOCUS)

	-- Get the GCD status
	local beforeNextAction
	local gcdStart, gcdDuration, gcdEnable = GetSpellCooldown(S.ArcaneShot)
	if gcdEnable == 1 and gcdStart and gcdDuration and gcdDuration <= GCD_DURATION then
		beforeNextAction = max(gcdStart + gcdDuration - now, 0)
	else
		beforeNextAction = 0
		lastCastSpell = nil -- the GCD of the last spell is done, forget it
	end

	-- Update target health upvalue
	targetHealthRatio = UnitHealth("target") / UnitHealthMax("target")

	-- Check current casting
	local _, startTime, endTime, prediction
	currentCast, _, _, _, startTime, endTime = UnitCastingInfo("player")
	currentCastEndTime = currentCast and max(startTime / 1000 + GCD_DURATION, endTime / 1000) or 0
	if currentCast then
		beforeNextAction = max(beforeNextAction, currentCastEndTime - now)
		prediction = focus - GetCost(currentCast)
	else
		prediction = focus
	end

	-- Update the FocusBar
	if self.FocusBar:IsVisible() then
		local nextActionFocus = prediction + EstimatePassiveIncome(beforeNextAction)
		self.FocusBar:SetValues(focus, focusMax)
		self.FillerBar:SetValues(nextActionFocus > focus and nextActionFocus or nil, focusMax)

		-- Update the position of spell marks
		local position = 0
		for i, name in ipairs(rotation.rotations[rotation.selector()]) do
			local mark = self.Marks[name]
			if mark then
				local cost, timeLeft, spell = calculate[name]()
				if cost and timeLeft then
					local delta = EstimatePassiveIncome(timeLeft)
					mark:SetPosition(position + cost - delta, timeLeft and timeLeft > 0 and timeLeft < 3 and format("%.1g", timeLeft) or nil, spell)
					position = position + max(0, cost - EstimatePassiveIncome(timeLeft + GetCastingTime(spell)))
				else
					mark:Hide()
				end
			end
		end
	end

	-- Update the spell forecast
	if self.Forecast:IsVisible() then
		self.Forecast:SetSpell(SpellForecast(prediction, beforeNextAction))
	end
end

function addon:OnUpdate(elapsed)
	updateTimer = updateTimer - elapsed
	if updateTimer <= 0 then
		updateTimer = updateTimer + 0.2
		self:Update("OnUpdate")
	end
end

function addon:SPELL_UPDATE_COOLDOWN()
	wipe(_cooldowns)
	for _, f in pairs(updaters.cooldowns) do
		f()
	end
	updateTimer = 0
end

function addon:UNIT_AURA(_, unit)
	if unit == "player" or unit == "target" or unit == "pet" then
		wipe(_auras)
		for _, f in pairs(updaters.auras) do
			f()
		end
		if unit == "player" then
			wipe(costs)
			for _, f in pairs(updaters.costs) do
				f()
			end
		end
		updateTimer = 0
	end
end

function addon:UNIT_STATS(_, unit)
	if unit == "player" then
		updateTimer = 0
	end
end

function addon:UNIT_POWER_FREQUENT(event, unit, powerType)
	if unit == "player" and powerType == "FOCUS" then
		updateTimer = 0
	end
end
addon.UNIT_MAXPOWER = addon.UNIT_POWER_FREQUENT

function addon:UNIT_HEALTH_FREQUENT(event, unit)
	if unit == 'target' then
		updateTimer = 0
	end
end

function addon:UNIT_SPELLCAST_SUCCEEDED(event, unit, _, _, _, spellId)
	if unit == "player" then
		if spellId == 75 then return end -- Ignore auto shots
		lastCastSpell, lastCastTime = spellId, GetTime()
		if spellId == S.SteadyShot then -- Steady Shot
			steadiesInRow = steadiesInRow + 1
		else
			steadiesInRow = 0
		end
		updateTimer = 0
	end
end

function addon:UNIT_SPELLCAST_START(event, unit, _, _, _, spellId)
	if unit == "player" and (spellId == S.CobraShot or spellId == S.SteadyShot) then
		updateTimer = 0
	end
end
addon.UNIT_SPELLCAST_STOP = addon.UNIT_SPELLCAST_START
